@@ -8,9 +8,13 @@ use base 'Exporter';
 use Carp;
 use Socket;
 use Config;
-use Math::BigInt lib => 'GMP';
+use Math::BigInt;
 use IO::Socket::INET;
 use IO::Socket::UNIX;
+use Encode qw/encode_utf8 decode_utf8/;
+
+my $is_native64 = $Config{longsize} == 8 || defined $Config{use64bitint} || defined $Config{use64bitall};
+    
 
 =head1 NAME
 
@@ -20,10 +24,11 @@ Sphinx::Search - Sphinx search engine API Perl client
 
 Please note that you *MUST* install a version which is compatible with your version of Sphinx.
 
-This version is 0.13_01, development release for Sphinx 0.9.9-rc1.  Open/Close
-for persistent connections are known to not work; also 64 bit IDs fail.
+Use version 0.23_02 for Sphinx svn-2269 (experimental)
 
-Use version 0.13 for Sphinx 0.9.9-rc1 and later
+Use version 0.22 for Sphinx 0.9.9-rc2 and later (Please read the Compatibility Note under L<SetEncoders> regarding encoding changes)
+
+Use version 0.15 for Sphinx 0.9.9-svn-r1674
 
 Use version 0.12 for Sphinx 0.9.8
 
@@ -43,17 +48,17 @@ Use version 0.02 for Sphinx 0.9.8-cvs-20070818
 
 =cut
 
-our $VERSION = '0.13_01';
+our $VERSION = '0.23_02';
 
 =head1 SYNOPSIS
 
     use Sphinx::Search;
 
-    $sphinx = Sphinx::Search->new();
+    $sph = Sphinx::Search->new();
 
-    $results = $sphinx->SetMatchMode(SPH_MATCH_ALL)
-                      ->SetSortMode(SPH_SORT_RELEVANCE)
-                      ->Query("search terms");
+    $results = $sph->SetMatchMode(SPH_MATCH_ALL)
+                   ->SetSortMode(SPH_SORT_RELEVANCE)
+                   ->Query("search terms");
 
 =head1 DESCRIPTION
 
@@ -79,12 +84,17 @@ use constant SEARCHD_COMMAND_EXCERPT	=> 1;
 use constant SEARCHD_COMMAND_UPDATE	=> 2;
 use constant SEARCHD_COMMAND_KEYWORDS	=> 3;
 use constant SEARCHD_COMMAND_PERSIST	=> 4;
+use constant SEARCHD_COMMAND_STATUS	=> 5;
+use constant SEARCHD_COMMAND_QUERY	=> 6;
+use constant SEARCHD_COMMAND_FLUSHATTRS	=> 7;
 
 # current client-side command implementation versions
-use constant VER_COMMAND_SEARCH		=> 0x116;
-use constant VER_COMMAND_EXCERPT	=> 0x100;
+use constant VER_COMMAND_SEARCH		=> 0x117;
+use constant VER_COMMAND_EXCERPT	=> 0x101;
 use constant VER_COMMAND_UPDATE	        => 0x102;
 use constant VER_COMMAND_KEYWORDS       => 0x100;
+use constant VER_COMMAND_STATUS         => 0x100;
+use constant VER_COMMAND_FLUSHATTRS     => 0x100;
 
 # known searchd status codes
 use constant SEARCHD_OK			=> 0;
@@ -108,6 +118,9 @@ use constant SPH_RANK_NONE              => 2; # no ranking, all matches get a we
 use constant SPH_RANK_WORDCOUNT         => 3; # simple word-count weighting, rank is a weighted sum of per-field keyword occurence counts
 use constant SPH_RANK_PROXIMITY         => 4;
 use constant SPH_RANK_MATCHANY          => 5;
+use constant SPH_RANK_FIELDMASK         => 6;
+use constant SPH_RANK_SPH04             => 7;
+use constant SPH_RANK_TOTAL             => 8;
 
 # known sort modes
 use constant SPH_SORT_RELEVANCE		=> 0;
@@ -129,6 +142,7 @@ use constant SPH_ATTR_ORDINAL		=> 3;
 use constant SPH_ATTR_BOOL		=> 4;
 use constant SPH_ATTR_FLOAT		=> 5;
 use constant SPH_ATTR_BIGINT		=> 6;
+use constant SPH_ATTR_STRING		=> 7;
 use constant SPH_ATTR_MULTI		=> 0x40000000;
 
 # known grouping functions
@@ -148,16 +162,8 @@ sub _sphPackI64 {
     my $v = shift;
 
     # x64 route
-    if ( $self->{_longsize} >= 8 ) {
-	my $i = int($v);
-	return pack ( "NN", $i>>32, $i & 4294967295 );
-    }
-
-    # x32 route, BigInt
-    my $vb = Math::BigInt->new($v);
-    my $l = $vb->copy->band ( 4294967295 );
-    my $h = $vb->copy->brsft ( 32 );
-    return pack ( "NN", $h, $l );
+    my $i = $is_native64 ? int($v) : Math::BigInt->new("$v");
+    return pack ( "NN", $i>>32, $i & 4294967295 );
 }
 
 # portably pack numeric to 64 unsigned bits, network order
@@ -165,17 +171,8 @@ sub _sphPackU64 {
     my $self = shift;
     my $v = shift;
 
-    # x64 route
-    if ( $self->{_longsize} >= 8 ) {
-	my $i = int($v);
-	return pack ( "NN", $i>>32, $i & 4294967295 );
-    }
-
-    # x32 route, BigInt
-    my $vb = Math::BigInt->new($v);
-    my $l = $vb->copy->band ( 4294967295 );
-    my $h = $vb->copy->brsft ( 32 );
-    return pack ( "NN", $h, $l );
+    my $i = $is_native64 ? int($v) : Math::BigInt->new("$v");
+    return pack ( "NN", $i>>32, $i & 4294967295 );
 }
 
 sub _sphPackI64array {
@@ -196,7 +193,7 @@ sub _sphUnpackU64
     my ($h,$l) = unpack ( "N*N*", $v );
 
     # x64 route
-    return ($h<<32) + $l if ( $self->{_longsize} >= 8 );
+    return ($h<<32) + $l if $is_native64;
 
     # x32 route, BigInt
     $h = Math::BigInt->new($h);
@@ -216,7 +213,7 @@ sub _sphUnpackI64
     my $neg = ($h & 0x80000000) ? 1 : 0;
 
     # x64 route
-    if ( $self->{_longsize} >= 8 ) {
+    if ( $is_native64 ) {
 	return -(~(($h<<32) + $l) + 1) if $neg;
 	return ($h<<32) + $l;
     }
@@ -273,7 +270,7 @@ sub new {
     my $self = {
 	# per=client-object settings
 	_host		=> 'localhost',
-	_port		=> 3312,
+	_port		=> 9312,
 	_path           => undef,
 	_socket         => undef,
 
@@ -285,7 +282,7 @@ sub new {
 	_sort		=> SPH_SORT_RELEVANCE,
 	_sortby		=> "",
 	_min_id		=> 0,
-	_max_id		=> 0xFFFFFFFF,
+	_max_id		=> 0,
 	_filters	=> [],
 	_groupby	=> "",
 	_groupdistinct	=> "",
@@ -306,19 +303,20 @@ sub new {
 	# per-reply fields (for single-query case)
 	_error		=> '',
 	_warning	=> '',
-	_connerror      => '',,
+	_connerror      => '',
 	
 	# request storage (for multi-query case)
 	_reqs           => [],
 	_timeout        => 0,
 
-	_longsize      => $Config{longsize},
+	_string_encoder => \&encode_utf8,
+	_string_decoder => \&decode_utf8,
     };
     bless $self, ref($class) || $class;
 
     # These options are supported in the constructor, but not recommended 
     # since there is no validation.  Use the Set* methods instead.
-    my %legal_opts = map { $_ => 1 } qw/host port offset limit mode weights sort sortby groupby groupbyfunc maxmatches cutoff retrycount retrydelay log debug/;
+    my %legal_opts = map { $_ => 1 } qw/host port offset limit mode weights sort sortby groupby groupbyfunc maxmatches cutoff retrycount retrydelay log debug string_encoder string_decoder/;
     for my $opt (keys %$options) {
 	$self->{'_' . $opt} = $options->{$opt} if $legal_opts{$opt};
     }
@@ -385,13 +383,60 @@ sub IsConnectError {
     return shift->{_connerror};
 }
 
+=head2 SetEncoders
+
+    $sph->SetEncoders(\&encode_function, \&decode_function)
+
+COMPATIBILITY NOTE: SetEncoders() was introduced in version 0.17.
+Prior to that, all strings were considered to be sequences of bytes
+which may have led to issues with multi-byte characters.  If you were
+previously encoding/decoding strings external to Sphinx::Search, you
+will need to disable encoding/decoding by setting Sphinx::Search to
+use raw values as explained below (or modify your code and let
+Sphinx::Search do the recoding).
+
+Set the string encoder/decoder functions for transferring strings
+between perl and Sphinx.  The encoder should take the perl internal
+representation and convert to the bytestream that searchd expects, and
+the decoder should take the bytestream returned by searchd and convert to
+perl format.
+
+The searchd format will depend on the 'charset_type' index setting in
+the Sphinx configuration file.
+
+The coders default to encode_utf8 and decode_utf8 respectively, which
+are compatible with the 'utf8' charset_type.
+
+If either the encoder or decoder functions are left undefined in the
+call to SetEncoders, they return to their default values.  
+
+If you wish to send raw values (no encoding/decoding), supply a
+function that simply returns its argument, e.g. 
+
+    $sph->SetEncoders( sub { shift }, sub { shift });
+
+Returns $sph.
+
+=cut
+
+sub SetEncoders {
+    my $self = shift;
+    my $encoder = shift;
+    my $decoder = shift;
+
+    $self->{_string_encoder} = $encoder ? $encoder : \&encode_utf8;
+    $self->{_string_decoder} = $decoder ? $decoder : \&decode_utf8;
+	
+    return $self;
+}
+
 =head2 SetServer
 
     $sph->SetServer($host, $port);
     $sph->SetServer($path, $port);
 
 In the first form, sets the host (string) and port (integer) details for the
-searchd server using a network (INET) socket.
+searchd server using a network (INET) socket (default is localhost:9312).
 
 In the second form, where $path is a local filesystem path (optionally prefixed
 by 'unix://'), sets the client to access the searchd server via a local (UNIX
@@ -457,7 +502,12 @@ sub _Send {
 sub _Connect {
 	my $self = shift;
 	
-	return $self->{_socket} if $self->{_socket};
+	if ($self->{_socket}) {
+	    # persistent connection, check it
+	    return $self->{_socket} if $self->{_socket}->connected;
+	    # force reopen
+	    undef $self->{_socket};
+	}
 
 	my $debug = $self->{_debug};
 	my $str_dest = $self->{_path} ? 'unix://' . $self->{_path} : "$self->{_host}:$self->{_port}";
@@ -531,9 +581,12 @@ sub _GetResponse {
 	my ($status, $ver, $len ) = unpack("n2N", $header);
         my $response = q{};
 	my $lasterror = q{};
+	my $lentotal = 0;
 	while (my $rlen = $fp->read(my $chunk, $len)) {
 	    $lasterror = $!, last if $rlen < 0;
 	    $response .= $chunk;
+	    $lentotal += $rlen;
+	    last if $lentotal >= $len;
 	}
         close($fp) unless $self->{_socket};
 
@@ -706,11 +759,8 @@ sub SetRankingMode {
     my $self = shift;
     my $ranker = shift;
 
-    croak("Unknown ranking mode: $ranker") unless ( $ranker==SPH_RANK_PROXIMITY_BM25
-						    || $ranker==SPH_RANK_BM25
-						    || $ranker==SPH_RANK_NONE
-						    || $ranker==SPH_RANK_WORDCOUNT
-						    || $ranker==SPH_RANK_PROXIMITY );
+    croak("Unknown ranking mode: $ranker") unless ( $ranker >= 0
+						    && $ranker < SPH_RANK_TOTAL );
 
     $self->{_ranker} = $ranker;
     return $self;
@@ -870,7 +920,7 @@ sub SetIDRange {
     $sph->SetFilter($attr, \@values, $exclude);
 
 Sets the results to be filtered on the given attribute.  Only results which have
-attributes matching the given (numeric) values will be returned.
+attributes matching the given values will be returned.
 
 This may be called multiple times with different attributes to select on
 multiple attributes.
@@ -888,9 +938,6 @@ sub SetFilter {
     croak("values is not an array reference") unless (ref($values) eq 'ARRAY');
     croak("values reference is empty") unless (scalar(@$values));
 
-    foreach my $value (@$values) {
-	croak("value $value is not numeric") unless ($value =~ m/$num_re/);
-    }
     push(@{$self->{_filters}}, {
 	type => SPH_FILTER_VALUES,
 	attr => $attribute,
@@ -1339,7 +1386,7 @@ sub AddQuery {
     my $req;
     $req = pack ( "NNNNN", $self->{_offset}, $self->{_limit}, $self->{_mode}, $self->{_ranker}, $self->{_sort} ); # mode and limits
     $req .= pack ( "N/a*", $self->{_sortby});
-    $req .= pack ( "N/a*", $query ); # query itself
+    $req .= pack ( "N/a*", $self->{_string_encoder}->($query) ); # query itself
     $req .= pack ( "N*", scalar(@{$self->{_weights}}), @{$self->{_weights}});
     $req .= pack ( "N/a*", $index); # indexes
     $req .= pack ( "N", 1) 
@@ -1464,7 +1511,7 @@ sub RunQueries {
 	return;
     }
 
-    my $fp = $self->_Connect() or return;
+    my $fp = $self->_Connect() or do { $self->{_reqs} = []; return };
 
     ##################
     # send query, get response
@@ -1563,6 +1610,10 @@ sub RunQueries {
 			push(@{$data->{$attr}}, $val);
 		    }
 		}
+		elsif ($attrs{$attr} == SPH_ATTR_STRING) {
+		    $data->{$attr} = $self->{_string_decoder}->(substr ($response, $p, $val));
+		    $p += $val;
+		}
 		else {
 		    $data->{$attr} = $val;
 		}
@@ -1577,7 +1628,7 @@ sub RunQueries {
 	while ( $words-->0 && $p < $max) {
 	    my $len = unpack ( "N*", substr ( $response, $p, 4 ) ); 
 	    $p += 4;
-	    my $word = substr ( $response, $p, $len ); 
+	    my $word = $self->{_string_decoder}->( substr ( $response, $p, $len ) ); 
 	    $p += $len;
 	    my ($docs, $hits) = unpack ("N*N*", substr($response, $p, 8));
 	    $p += 8;
@@ -1637,13 +1688,17 @@ A hash which contains additional optional highlighting parameters:
 
 =item weight_order 
 
+=item query_mode
+
+=item force_all_words
+
 =back
 
 =back
 
 Returns undef on failure.
 
-Returns an array of string excerpts on success.
+Returns an array ref of string excerpts on success.
 
 =cut
 
@@ -1669,22 +1724,26 @@ sub BuildExcerpts {
 	$opts->{"single_passage"} ||= 0;
 	$opts->{"use_boundaries"} ||= 0;
 	$opts->{"weight_order"} ||= 0;
+	$opts->{"query_mode"} ||= 0;
+	$opts->{"force_all_words"} ||= 0;
 
 	##################
 	# build request
 	##################
 
-	# v.1.0 req
+	# v.1.01 req
 	my $req;
 	my $flags = 1; # remove spaces
 	$flags |= 2 if ( $opts->{"exact_phrase"} );
 	$flags |= 4 if ( $opts->{"single_passage"} );
 	$flags |= 8 if ( $opts->{"use_boundaries"} );
 	$flags |= 16 if ( $opts->{"weight_order"} );
+	$flags |= 32 if ( $opts->{"query_mode"} );
+	$flags |= 64 if ( $opts->{"force_all_words"} );
 	$req = pack ( "NN", 0, $flags ); # mode=0, flags=$flags
 
-	$req .= pack ( "N", length($index) ) . $index; # req index
-	$req .= pack ( "N", length($words) ) . $words; # req words
+	$req .= pack ( "N/a*", $index ); # req index
+	$req .= pack ( "N/a*", $self->{_string_encoder}->($words)); # req words
 
 	# options
 	$req .= pack ( "N/a*", $opts->{"before_match"});
@@ -1696,8 +1755,8 @@ sub BuildExcerpts {
 	# documents
 	$req .= pack ( "N", scalar(@$docs) );
 	foreach my $doc (@$docs) {
-		croak('BuildExcepts: Found empty document in $docs') unless ($doc);
-		$req .= pack("N/a*", $doc);
+		croak('BuildExcerpts: Found empty document in $docs') unless ($doc);
+		$req .= pack("N/a*", $self->{_string_encoder}->($doc));
 	}
 
 	##########################
@@ -1708,7 +1767,7 @@ sub BuildExcerpts {
 	$self->_Send($fp, $req);
 	
 	my $response = $self->_GetResponse($fp, VER_COMMAND_EXCERPT);
-	return 0 unless ($response);
+	return unless $response;
 
 	my ($pos, $i) = 0;
 	my $res = [];	# Empty hash ref
@@ -1719,9 +1778,9 @@ sub BuildExcerpts {
 
                 if ( $pos+$len > $rlen ) {
 			$self->_Error("incomplete reply");
-			return 0;
+			return;
 		}
-		push(@$res, substr ( $response, $pos, $len ));
+		push(@$res, $self->{_string_decoder}->( substr ( $response, $pos, $len ) ));
 		$pos += $len;
         }
         return $res;
@@ -1764,9 +1823,9 @@ sub BuildKeywords {
     my $fp = $self->_Connect() or return;
 
     # v.1.0 req
-    my $req = pack("N/a*", $query);
+    my $req = pack("N/a*", $self->{_string_encoder}->($query) );
     $req .= pack("N/a*", $index);
-    $req .= pack("N", $hits);
+    $req .= pack("N", $self->{_string_encoder}->($hits) );
 
     ##################
     # send query, get response
@@ -1790,10 +1849,10 @@ sub BuildKeywords {
     for (my $i=0; $i < $nwords; $i++ ) {
 	my $len = unpack("N", substr ( $response, $p, 4 ) ); $p += 4;
 
-	my $tokenized = $len ? substr ( $response, $p, $len ) : ""; $p += $len;
+	my $tokenized = $len ? $self->{_string_decoder}->( substr ( $response, $p, $len ) ) : ""; $p += $len;
 	$len = unpack("N", substr ( $response, $p, 4 ) ); $p += 4;
 
-	my $normalized = $len ? substr ( $response, $p, $len ) : ""; $p += $len;
+	my $normalized = $len ? $self->{_string_decoder}->( substr ( $response, $p, $len ) ) : ""; $p += $len;
 	my %data = ( tokenized => $tokenized, normalized => $normalized );
 	
 	if ($hits) {
@@ -1981,6 +2040,72 @@ sub Close {
     return 1;
 }
 
+=head2 Status
+
+    $status = $sph->Status()
+
+Queries searchd status, and returns a hash of status variable name and value pairs. 
+
+Returns undef on failure.
+
+=cut
+
+sub Status {
+    
+    my $self = shift;
+
+    my $fp = $self->_Connect() or return;
+   
+    my $req = pack("nnNN", SEARCHD_COMMAND_STATUS, VER_COMMAND_STATUS, 4, 1 ); # len=4, body=1
+    $self->_Send($fp, $req) or return;
+    my $response = $self->_GetResponse ( $fp, VER_COMMAND_STATUS );
+    return unless $response;
+
+    my $p = 0;
+    my ($rows, $cols) = unpack("N*N*", substr ( $response, $p, 8 ) ); $p += 8;
+
+    return {} unless $rows && $cols;
+    my %res;
+    for (1 .. $rows ) {
+	my @entry;
+	for ( 1 .. $cols) {
+	    my $len = unpack("N*", substr ( $response, $p, 4 ) ); $p += 4;
+	    push(@entry, $len ? substr ( $response, $p, $len ) : ""); $p += $len;
+	}
+	if ($cols <= 2) {
+	    $res{$entry[0]} = $entry[1];
+	}
+	else {
+	    my $name = shift @entry;
+	    $res{$name} = \@entry;
+	}
+    }
+    return \%res;
+}
+    
+=head2 FlushAttributes
+
+=cut
+
+sub FlushAttributes {
+    my $self = shift;
+
+    my $fp = $self->_Connect() or return;
+    
+    my $req = pack("nnN", SEARCHD_COMMAND_FLUSHATTRS, VER_COMMAND_FLUSHATTRS, 0 ); # len=0
+    $self->_Send($fp, $req) or return;
+    my $response = $self->_GetResponse ( $fp, VER_COMMAND_FLUSHATTRS );
+    return unless $response;
+    
+    my $tag = -1;
+    if (length($response) == 4) {
+	$tag = unpack ( "N*", substr ( $response, 0, 4 ) );
+    }
+    else {
+	$self->_Error("unexpected response length");
+    }
+    return $tag;
+}    
 
 =head1 SEE ALSO
 
@@ -1988,7 +2113,7 @@ L<http://www.sphinxsearch.com>
 
 =head1 NOTES
 
-There is a bundled Sphinx.pm in the contrib area of the Sphinx source
+There is (or was) a bundled Sphinx.pm in the contrib area of the Sphinx source
 distribution, which was used as the starting point of Sphinx::Search.
 Maintenance of that version appears to have lapsed at sphinx-0.9.7, so many of
 the newer API calls are not available there.  Sphinx::Search is mostly
